/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.jbi.messaging.exchange;

import java.net.URI;
import java.util.logging.Logger;

import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.InOptionalOut;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.RobustInOnly;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfOperation.MEPPatternConstants;
import org.ow2.petals.jbi.messaging.exchange.impl.DefaultMessageExchangeFactory;
import org.ow2.petals.jbi.messaging.exchange.types.InOnlyImpl;
import org.ow2.petals.jbi.messaging.exchange.types.InOptionalOutImpl;
import org.ow2.petals.jbi.messaging.exchange.types.InOutImpl;
import org.ow2.petals.jbi.messaging.exchange.types.RobustInOnlyImpl;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This implementation creates {@link MessageExchangeWrapper} instances of {@link MessageExchange}.
 * @author Olivier Fabre - EBM WebSourcing
 */
public abstract class AbstractMessageExchangeFactory implements MessageExchangeFactory {

    protected final LoggingUtil log;

    /**
     * The reply address for any service request
     */
    private final ServiceEndpoint consumerEndpoint;

    private final ServiceEndpoint defaultEndpoint;

    private final QName defaultInterfaceName;

    private final QName defaultServiceName;

    protected AbstractMessageExchangeFactory(final ServiceEndpoint consumerEndpoint,
            final ServiceEndpoint targetEndpoint, final QName targetInterface, final QName targetService,
            final Logger logger) {
        assert consumerEndpoint != null;
        this.consumerEndpoint = consumerEndpoint;
        this.defaultEndpoint = targetEndpoint;
        this.defaultInterfaceName = targetInterface;
        this.defaultServiceName = targetService;
        this.log = new LoggingUtil(logger);
    }

    @Override
    public InOnly createInOnlyExchange() throws MessagingException {
        this.log.call();

        return (InOnly) buildExchangeDecorator(MEPPatternConstants.IN_ONLY.value());
    }

    @Override
    public InOptionalOut createInOptionalOutExchange() throws MessagingException {
        this.log.call();

        return (InOptionalOut) buildExchangeDecorator(MEPPatternConstants.IN_OPTIONAL_OUT.value());
    }

    @Override
    public InOut createInOutExchange() throws MessagingException {
        this.log.call();

        return (InOut) buildExchangeDecorator(MEPPatternConstants.IN_OUT.value());
    }

    @Override
    public RobustInOnly createRobustInOnlyExchange() throws MessagingException {
        this.log.call();

        return (RobustInOnly) buildExchangeDecorator(MEPPatternConstants.ROBUST_IN_ONLY.value());
    }

    @Override
    public MessageExchange createExchange(final QName serviceName, QName operationName)
            throws MessagingException {
        this.log.call("Service Name : " + serviceName + " - operationName : " + operationName);
        assert serviceName != null;
        assert operationName != null;

        final MessageExchangeWrapper messageExchangeDecorator = buildExchangeDecorator(null);

        messageExchangeDecorator.setService(serviceName);
        messageExchangeDecorator.setOperation(operationName);

        return messageExchangeDecorator;
    }

    @Override
    public MessageExchange createExchange(final URI pattern) throws MessagingException {
        this.log.call(pattern);
        assert pattern != null;

        checkPatternIsJBISupportedPattern(pattern);

        return buildExchangeDecorator(pattern);
    }

    private void checkPatternIsJBISupportedPattern(final URI pattern) throws MessagingException {
        if (!MEPPatternConstants.IN_ONLY.equals(pattern) && !MEPPatternConstants.ROBUST_IN_ONLY.equals(pattern)
                && !MEPPatternConstants.IN_OUT.equals(pattern) && !MEPPatternConstants.IN_OPTIONAL_OUT.equals(pattern)) {
            throw new MessagingException("This Message Exchange Pattern is not recognized by JBI.");
        }
    }

    private MessageExchangeWrapper buildExchangeDecorator(final URI pattern) throws MessagingException {
        this.log.call(pattern);

        final PetalsMessageExchange me = createExchange();
        me.setPattern(pattern);
        return createMessageExchangeDecoratorForPattern(me);
    }

    /**
     * This is responsible of building and configuring the exchange
     * 
     * @return
     */
    protected PetalsMessageExchange createExchange() {
        final PetalsMessageExchange msg = buildExchange();

        msg.setConsumerEndpoint(consumerEndpoint);
        msg.setEndpoint(defaultEndpoint);
        msg.setInterfaceName(defaultInterfaceName);
        msg.setService(defaultServiceName);

        return msg;
    }

    /**
     * Override to create a new {@link PetalsMessageExchange}.
     * 
     * @see DefaultMessageExchangeFactory for an example.
     */
    protected abstract PetalsMessageExchange buildExchange();

    /**
     * Override, if needed, to customize
     */
    protected MessageExchangeWrapper createMessageExchangeDecoratorForPattern(final PetalsMessageExchange exchange) {
        final URI pattern = exchange.getPattern();
        if (isInOnlyPattern(pattern)) {
            return new InOnlyImpl(exchange);
        } else if (isRobustInOnlyPattern(pattern)) {
            return new RobustInOnlyImpl(exchange);
        } else if (isInOutPattern(pattern)) {
            return new InOutImpl(exchange);
        } else if (isInOptionalOutPattern(pattern)) {
            return new InOptionalOutImpl(exchange);
        } else {
            if (pattern != null) {
                this.log.warning("Creating an exchange with unknown pattern: " + pattern);
            }
            return new MessageExchangeWrapper(exchange);
        }
    }

    protected boolean isInOptionalOutPattern(final URI pattern) {
        return MEPPatternConstants.IN_OPTIONAL_OUT.equals(pattern);
    }

    protected boolean isInOutPattern(final URI pattern) {
        return MEPPatternConstants.IN_OUT.equals(pattern);
    }

    protected boolean isRobustInOnlyPattern(final URI pattern) {
        return MEPPatternConstants.ROBUST_IN_ONLY.equals(pattern);
    }

    protected boolean isInOnlyPattern(final URI pattern) {
        return MEPPatternConstants.IN_ONLY.equals(pattern);
    }
}
